using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;

/// <summary>

/// </summary>


public class VaginaSoftPhysics : MVRScript
{
	DAZPhysicsMesh lowerMesh;
	DAZPhysicsMeshSoftVerticesGroup genGroup;
	
	DAZPhysicsMeshSoftVerticesGroup leftGroup;
	DAZPhysicsMeshSoftVerticesGroup rightGroup;
	
	bool isSessionPlugin = false;
	
	public override void Init() {
		// Might be a better way to do this, but works for now
		isSessionPlugin = containingAtom.name == "CoreControl";
		
		if(isSessionPlugin){
			StartCoroutine("AddSets");
			return;
		} else {
			SuperController.LogError("This plugin only works as a session plugin and must be loaded before any models.");
		}
	}
	
	private IEnumerator AddSets(){
		// Check to see if the prefab already exists (there's a person in the scene)
		GameObject personPrefab = SuperController.singleton.GetCachedPrefab("a_per", "Person");
		Atom prefabAtom = null;
		if(personPrefab == null){
			// No people in the scene yet. Load one (without adding to scene) and register it as the prefab. 
			// This might have unintended consequences. 
			SuperController.LogMessage("Person prefab not cached. Loading from asset bundle.");
			AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_per", "Person", typeof(GameObject));
			if (request == null)
			{
				SuperController.LogError("Unable to load Person prefab");
				yield break;
			}
			yield return base.StartCoroutine(request);

			personPrefab = request.GetAsset<GameObject>();
			if (personPrefab != null)
			{
				SuperController.singleton.RegisterPrefab("a_per", "Person", personPrefab);
				prefabAtom = personPrefab.GetComponent<Atom>();
			} else {
				SuperController.LogError("Unable to load Person prefab");
				yield break;
			}
		} else {
			prefabAtom = personPrefab.GetComponent<Atom>();
		}
		
		foreach(DAZPhysicsMesh mesh in prefabAtom.GetComponentsInChildren<DAZPhysicsMesh>()){
			if(mesh.name.Contains("Lower")){
				foreach(DAZPhysicsMeshSoftVerticesGroup vertGroup in mesh.softVerticesGroups){
					if(vertGroup.name == "gen"){
						lowerMesh = mesh;
						genGroup = vertGroup;
					}
					
					// The existing soft physics groups influence some vertices that we want 
					// to add physics for, so we're going remove them
					foreach(var vertSet in vertGroup.softVerticesSets){
						List<int> toRemove = new List<int>();
						for(int i = 0; i < vertSet.influenceVertices.Length; i++){
							int index = vertSet.influenceVertices[i];
							if(index >= 22086 && index <= 22197){
								toRemove.Add(index);
							}
							
							if(index >= 22676 && index <= 22788){
								toRemove.Add(index);
							}
						}
						foreach(int index in toRemove){
							vertSet.RemoveInfluenceVertex(index);
						}
					}
				}
			}
		}

		// Add some new pysics groups
		leftGroup = InitGroup(lowerMesh, "genLeft", 22086, 22197);
		rightGroup = InitGroup(lowerMesh, "genRight", 22676, 22788);
		
		
		// Set some physics constants
		// TODO: add some sliders?
		leftGroup.linkSpring = 10000f;
		rightGroup.linkSpring = 10000f;
		
		leftGroup.linkDamper = 0f;
		rightGroup.linkDamper = 0f;
		// up and down
		leftGroup.normalDistanceLimit = 0.001f;
		leftGroup.normalNegativeDistanceLimit = 0.001f;
		rightGroup.normalDistanceLimit = 0.001f;
		rightGroup.normalNegativeDistanceLimit = 0.001f;
		
		// forward and back
		leftGroup.tangentDistanceLimit = 0.01f;
		leftGroup.tangentNegativeDistanceLimit = 0.01f;
		rightGroup.tangentDistanceLimit = 0.01f;
		rightGroup.tangentNegativeDistanceLimit = 0.01f;
		
		// left and right
		leftGroup.tangent2DistanceLimit = 0.0f;
		leftGroup.tangent2NegativeDistanceLimit = 10.0f;
		rightGroup.tangent2DistanceLimit = 10.0f;
		rightGroup.tangent2NegativeDistanceLimit = 0.0f;
		
		SuperController.LogMessage($"Successfully added new soft physics groups");
	}
	
	public DAZPhysicsMeshSoftVerticesGroup InitGroup(DAZPhysicsMesh parent, string name, int startVertIndex, int endVertIndex){
		int newGroupIndex = parent.AddSoftVerticesGroup();
		var newGroup = lowerMesh.softVerticesGroups[newGroupIndex];
		newGroup.name = name;
		
		
		// For now we'll only create groups that influence one vertex. 
		// This is more compute-intensive but the vaginal canal doesn't
		// have too many vertices
		for(int i = startVertIndex; i <= endVertIndex; i++){
			int newSetIndex = newGroup.AddSet();
			var newSet = newGroup.softVerticesSets[newSetIndex];
			newSet.targetVertex = i;
			newSet.anchorVertex = i;
			newSet.influenceVertices = new int[]{};
		}
			
		foreach(var vertSet in newGroup.softVerticesSets){
			var setDistances = closestToVertex(vertSet, parent.skin.dazMesh.baseMesh.vertices, newGroup.softVerticesSets, 6);
			foreach(var setDistance in setDistances){
				var linkSet = setDistance.set;
				
				bool alreadyLinked = false;
				foreach(var otherLink in linkSet.links){
					if(otherLink == vertSet.uid){
						alreadyLinked = true;
						break;
					}
				}
				if(!alreadyLinked){
					vertSet.links.Add(linkSet.uid);
				}
			}
		}
				
		newGroup.normalReference = genGroup.normalReference;
		newGroup.ignoreColliders = genGroup.ignoreColliders;
		newGroup.ignoreAutoColliderGroups = genGroup.ignoreAutoColliderGroups;
		newGroup.colliderMaterial = genGroup.colliderMaterial;
		
		newGroup.centerBetweenTargetAndAnchor = false;
		newGroup.colliderType = DAZPhysicsMeshSoftVerticesGroup.ColliderType.Sphere;
		newGroup.lookAtOption = DAZPhysicsMeshSoftVerticesGroup.LookAtOption.NormalReference;
		
		newGroup.normalMovementType = DAZPhysicsMeshSoftVerticesGroup.MovementType.Limit;
		newGroup.tangentMovementType = DAZPhysicsMeshSoftVerticesGroup.MovementType.Limit;
		newGroup.tangent2MovementType = DAZPhysicsMeshSoftVerticesGroup.MovementType.Limit;
		
		
		return newGroup;
	}
	
	public class VertSetDistance {
		public DAZPhysicsMeshSoftVerticesSet set;
		public float distance;
	}
	
	public LinkedList<VertSetDistance> closestToVertex(DAZPhysicsMeshSoftVerticesSet baseSet, Vector3[] vertices, List<DAZPhysicsMeshSoftVerticesSet> sets, int max){
		LinkedList<VertSetDistance> list = new LinkedList<VertSetDistance>();
		
		Vector3 baseVertex = vertices[baseSet.targetVertex];
		
		foreach(var otherSet in sets){
			if(baseSet == otherSet) continue;
			
			Vector3 otherVertex = vertices[otherSet.targetVertex];
			
			float distance = (baseVertex - otherVertex).magnitude;
			VertSetDistance entry = new VertSetDistance();
			entry.set = otherSet;
			entry.distance = distance;
			
			InsertSorted(list, entry, max);
		}
		
		return list;
	}
	
	public static void InsertSorted(LinkedList<VertSetDistance> list, VertSetDistance value, int maxSize)
	{
		var newNode = new LinkedListNode<VertSetDistance>(value);

		if (list.First == null || value.distance < list.First.Value.distance)
		{
			list.AddFirst(newNode);
			return;
		}

		var current = list.First;
		while (current.Next != null && current.Next.Value.distance < value.distance)
		{
			current = current.Next;
		}

		list.AddAfter(current, newNode);
		
		if (list.Count > maxSize)
		{
			while (list.Count > maxSize)
			{
				list.RemoveLast();
			}
		}
	}
	
	private void Start()
	{
	}
	
	private void OnDisable()
	{
	}
}
